/* ptst.c - test the ports features */

#include <kernel.h>
#include <conf.h>
#include <dos.h>
#include <mem.h>
#include <mark.h>
#include <ports.h>

#define MM	15		/* maximum messages for this port */
#define TRIES	1000		/* no. of messages to send */
#define	SENDING		1
#define RECEIVING	2

int	marray[TRIES];		/* message array			*/
int	psem;			/* mutex sem. for printing status info	*/
int	done;			/* semaphore for exiting program	*/

/*------------------------------------------------------------------------
 *  xmain  --  initialize and start the port test processes
 *------------------------------------------------------------------------
 */
xmain()
{
	int	putem(),getem();
	int	pn;			/* port number			*/

	pn = pcreate(MM);		/* create the port		*/
	psem = screate(1);		/* synchronize status printing	*/
	done = screate(0);		/* synchronize program exit	*/
	resume(create(putem,INITSTK,INITPRIO,"putem",1,pn));
	resume(create(getem,INITSTK,INITPRIO,"getem",1,pn));
	wait(done);			/* wait for getem & putem	*/
	wait(done);
	printf("\nDone!\n");
	xdone();
}

/*------------------------------------------------------------------------
 *  putem  --  send a message address to the port
 *------------------------------------------------------------------------
 */
putem(pn)
int	pn;
{
	int	i;			/* contents of sent message	*/
	int	*mp;			/* message address itself	*/

	for ( i=0 ; i<TRIES ; i++ ) {
		mp = &marray[i];
		*mp = i;
		doing(SENDING,i,mp);
		if ( psend(pn,mp) == SYSERR )
			panic("psend failure");
	}
	signal(done);			/* done with sending		*/
}

/*------------------------------------------------------------------------
 *  getem  --  get a message address from the port
 *------------------------------------------------------------------------
 */
getem(pn)
int	pn;
{
	int	i;			/* contents of received message	*/
	int	*mp;			/* received message address	*/

	for ( i=0 ; i<TRIES ; i++ ) {
		if ( (mp=(int *)preceive(pn)) == (int *)NULL )
			panic("preceive failure");
		if ( i != *mp )
			panic("preceive inconsistency");
		doing(RECEIVING,i,mp);
	}
	signal(done);			/* done with receiving		*/
}

static int	last = 0;

/*------------------------------------------------------------------------
 *  doing  --  tell what we're doing at this point
 *------------------------------------------------------------------------
 */
doing(what,i,where)
int	what; /* what is happening now */
int	i; /* message number */
char *where; /* where it is going to or coming from (not used) */
{
	wait(psem);
	if ( what != last ) {
		switch ( what ) {
		case SENDING:
			printf("\nsending  ");
			break;
		case RECEIVING:
			printf("\nreceiving");
			break;
		}
		last = what;
	}
	printf(" %3d",i);		/* message sent/received	*/
	signal(psem);
}
